import logging
from datetime import datetime,timedelta
from logging.handlers import TimedRotatingFileHandler
from decimal import Decimal
from sqlalchemy.orm import Session

from .base import BaseNetworkClient

# from vc.
from vc.client.soldo import user, wallets, card, group, order, transaction
from vc.models.soldo import WalletSo, CardSo, TransactionSo
from vc.settings import Settings
from .soldo_event import EventMixer
from vc.libs.utils import set_config, generate_list_model_search_id
from decimal import Decimal

logger = logging.getLogger(__name__)
set_config(logger)


class SoldoException(Exception):

    def __init__(self, msg: str, response_data: dict = None):
        self.msg = msg
        self.response_data = response_data

    def __str__(self):
        return f"{self.msg}, {str(self.response_data)}"


def card_issue():
    return 7


class Soldo(EventMixer, BaseNetworkClient):
    settings = Settings({
        "ACCESS_TOKEN": "dFTPdcdQqHXI3GSJhPr5b2MAihDXGqWR",
    })
    event_list = ["new_user", "wallet_created", "store_order_completed"]
    __cache = {'0765a95e-6016-43fe-a1b4-4e66713e0a00':
                   {'wallet_id': 1, 'status': 'PLACED',
                    "category": "CARD"}, }
    __card_issue = None # function get card issue

    @property
    def card_price(self):
        return self.__card_issue()

    def get_cache(self):
        return self.__cache

    def get_cache_by_key(self, key):
        return self.__cache.get(key)

    def update_cache(self, **data: dict):
        return self.__cache.update(data)

    def remove_cache(self, key):
        if self.__cache.get(key):
            return self.__cache.pop(key)

    def __init__(self, name, uri,
                 api_url: str,
                 client_id: str,
                 client_secret: str,
                 group_id: str,
                 token: str,
                 safe_wallet: str,
                 safe_user: str,
                 filepath_private_pem: str,
                 card_issue=card_issue, log_file: str = None, currency="USD", user_model=None, **config):
        data = dict(name=name, currency=currency,
                    CLIENT_ID=client_id,
                    CLIENT_SECRET=client_secret,
                    API_URL=api_url,
                    TOKEN=token,
                    SAFE_WALLET=safe_wallet,
                    SAFE_USER=safe_user,
                    USER_MODEL=user_model,
                    UPLOAD_SIZE=1000500,
                    RULES=["OpenClose", "MaxPerTx"],
                    LOG_FILE=log_file,
                    PATH_RSA_PRIVATE=filepath_private_pem,
                    GROUP_ID=group_id, **config)
        Soldo.settings.update_config(**data)

        super().__init__(uri, user_model=user_model, **config)
        self.__card_issue = card_issue
        self.set_safe(safe_wallet, safe_user)

    def set_safe(self, wallet_id: str, user_email: str):

        wallet = wallets.get(wallet_id=wallet_id).data
        with self.get_session() as session:
            user = session.query(self._user).filter(self._user.email == user_email).first()

            wallet_q = session.query(WalletSo).filter(WalletSo.search_id==wallet_id).first()
            update_data = dict(
                user_id=user.id,
                search_id=wallet.id,
                balance=wallet.available_amount,
                blocked_balance=wallet.blocked_amount,
                currency=wallet.currency_code
            )
            if not wallet_q:
                wallet_q = WalletSo()
            for field in update_data:
                setattr(wallet_q, field, update_data[field])
            self.save_obj(session, obj=wallet_q)
        return wallet_q

    def activate_service(self, db: Session, user_id: int, owner_type="company"):
        user = db.query(self._user).filter(self._user.id == user_id).first()
        logger.debug(user.__dict__)
        response_data = wallets.create(owner_type, self.settings.currency, name=user.email, custom_reference_id=user.id)
        order = response_data.data
        if not order.is_valid or order.status != "PLACED":
            raise SoldoException("Error create_wallet", response_data.dict())

        wallet = WalletSo(user_id=user.id, currency=self.settings.currency, )
        if order.items:
            wallet.search_id = order.items[0].id

        self.save_obj(db, wallet)
        return wallet

    def get_transactions_by_wallet_id(self, db: Session, wallet_id, type="wallet", page=0,  **kwargs):
        wallet_q = db.query(WalletSo).filter(WalletSo.id == wallet_id).first()
        response_data = transaction.search(type=type, publicId=wallet_q.search_id, page=page, **kwargs).data
        return response_data

    def get_statements_by_card_id(self, db: Session, card_id, page=0, **kwargs):
        card_q = db.query(CardSo).filter(CardSo.id==card_id).first()
        response_data = transaction.search(type="card", publicId=card_q.search_id, page=page, **kwargs).data
        return response_data

    def update_card_rule(self, db: Session, card_id: int, name: str, enabled: bool, amount: Decimal = None):
        card_q = db.query(CardSo).filter(CardSo.id==card_id).first()
        if name not in self.settings.RULES:
            raise SoldoException("Exception rule. name rule not enable or not exists")
        rules = card.update_card_rule(card_q.search_id, name=name, enabled=enabled, amount=amount).data
        return rules

    def get_card_rules(self, db: Session, card_id: int):
        card_q = db.query(CardSo).filter(CardSo.id == card_id).first()
        return card.get_card_rules(card_q.search_id).data

    def upload_cards(self, db: Session):
        response_cards = card.search(page_size=self.settings.UPLOAD_SIZE).data.results
        query_wallets = db.query(WalletSo).filter(WalletSo.search_id is not None).all()
        query_cards = db.query(CardSo).filter(CardSo.search_id is not None).all()

        list_model_id, list_model = generate_list_model_search_id(query_cards)
        list_wallets_id, list_wallets = generate_list_model_search_id(query_wallets)

        filter_model = list(filter(
            lambda w: w.wallet_id in list_wallets_id,
            response_cards
        ))

        update_cards = []
        create_cards_id = []
        for model in filter_model:
            card_id = list_model.get(model.id)
            if card_id:
                update_cards.append({
                    "id": card_id,
                    "status": model.status,
                    "name": model.name
                })
            else:
                create_cards_id.append(model.id)

        create_cards = []
        for card_id in create_cards_id:
            print(card_id)
            response_card = card.get(card_id, True).data
            wallet_id = list_wallets.get(response_card.wallet_id)
            if wallet_id:
                print(response_card)
                card_obj = CardSo(wallet_id=wallet_id, search_id=response_card.id)
                card_obj.PAN = response_card.pan
                card_obj.cvv = response_card.cvv
                card_obj.name = response_card.name
                card_obj.type = response_card.type
                card_obj.created_on = response_card.creation_time
                card_obj.expiry = response_card.expiration_date.date()
                card_obj.PAN = response_card.pan
                card_obj.status = response_card.status
                create_cards.append(card_obj)

        db.bulk_update_mappings(CardSo, update_cards)
        if create_cards:
            db.bulk_save_objects(create_cards)
        db.commit()
        return {
            "create_list": [c.search_id for c in create_cards],
            "update_list": [c.get("id") for c in update_cards],
                }

    def internal_transfer(self, db: Session, from_wallet_id: int, to_wallet_id: int, amount: Decimal, currency: str ="EUR", **kwargs):
        from_wallet = db.query(WalletSo).filter(WalletSo.id==from_wallet_id).first()
        to_wallet = db.query(WalletSo).filter(WalletSo.id==to_wallet_id).first()
        response_data = wallets.internal_transfer(
            amount=amount,
            toWalletId=to_wallet.search_id,
            fromWalletId=from_wallet.search_id,
            currencyCode=currency,
            **kwargs
        ).data
        return response_data

    def internal_transfer_to_safe(self, db: Session, from_wallet_id: int, amount: Decimal,
                          currency: str = "EUR"):
        from_wallet = db.query(WalletSo).filter(WalletSo.id==from_wallet_id).first()
        to_wallet = db.query(WalletSo).filter(WalletSo.search_id==self.settings.SAFE_WALLET).first()
        response_data = wallets.internal_transfer(
            amount=amount,
            toWalletId=to_wallet.search_id,
            fromWalletId=from_wallet.search_id,
            currencyCode=currency
        ).data
        return response_data

    def upload_transaction_wallet(self, db: Session, **kwargs):
        date_to = datetime.now()
        date_from = date_to - timedelta(days=2)
        query_transactions = db.query(TransactionSo).filter(
            TransactionSo.date >= date_from,
            TransactionSo.date <= date_to + timedelta(days=1),
        ).all()
        query_wallets = db.query(WalletSo).filter(WalletSo.search_id is not None).all()
        print(query_transactions)

        list_trans_id, list_trans = generate_list_model_search_id(query_transactions)
        list_wallet_id, list_wallet = generate_list_model_search_id(query_wallets)

        response_data = transaction.search(
            page_size=self.settings.UPLOAD_SIZE,
            toDate=date_to.strftime(self.format_date),
            fromDate=date_from.strftime(self.format_date), **kwargs).data
        # upload safe wallet
        response_data += transaction.search(
            page_size=self.settings.UPLOAD_SIZE,
            toDate=date_to.strftime(self.format_date),
            fromDate=date_from.strftime(self.format_date),
            type="main", publicId=self.settings.SAFE_WALLET,
        ).data
        # filter wallet_id
        trans_by_wallet_id = list(filter(
            lambda trans: trans.wallet_id in list_wallet_id,
            response_data.results
        ))
        print(trans_by_wallet_id)
        # return

        create_transaction = []
        update_transaction = []
        columns = [c.name for c in TransactionSo.__table__.columns]
        columns.remove("id")
        for trans in trans_by_wallet_id:
            data = {}
            for c in columns:
                data[c] = getattr(trans, c, None)
            data["search_id"] = trans.id
            data['wallet_id'] = list_wallet.get(trans.wallet_id)

            trans_id  = list_trans.get(trans.id)
            if trans_id:
                data.update(dict(
                    id=trans_id,
                    search_id=trans.id
                ))
                update_transaction.append(data)
            else:
                create_transaction.append(
                    TransactionSo(
                        **data,
                    ))


        db.bulk_update_mappings(TransactionSo, update_transaction)
        if create_transaction:
            db.bulk_save_objects(create_transaction)
        db.commit()
        return {
            "create_list": [c.search_id for c in create_transaction],
            "update_list": [c.get("id") for c in update_transaction],
                }

    def upload_wallets(self, db: Session):
        response_wallets = wallets.search(type="company", page_size=self.settings.UPLOAD_SIZE).data.results
        query_wallets = db.query(WalletSo).filter(WalletSo.search_id is not None).all()

        list_wallets_id, list_wallets = generate_list_model_search_id(query_wallets)
        filter_wallet = list(filter(
            lambda w: w.id in list_wallets_id,
            response_wallets
        ))
        update_wallets = []
        for wallet in filter_wallet:
            wallet_id = list_wallets.get(wallet.id)
            update_wallets.append({
                    "id": wallet_id,
                    "balance": wallet.available_amount,
                    "blocked_balance": wallet.blocked_amount,
                    "currency": wallet.currency_code
                })

        db.bulk_update_mappings(WalletSo, update_wallets)
        db.commit()
        return list_wallets

    def wallet_update_balance(self, db: Session, wallet_id: int):
        wallet = db.query(WalletSo).filter(WalletSo.id == wallet_id).first()
        response_data = wallets.get(wallet.search_id)
        wallet.balance = response_data.data.available_amount
        wallet.blocked_balance = response_data.data.blocked_amount
        self.save_obj(db, wallet)
        return wallet

    def get_wallets(self, **kwargs):
        response_data = wallets.search(**kwargs)
        print(response_data)
        return response_data

    def get_card(self, card_id: str = None, showSensitiveData: str = None, **kwargs):
        return card.get(card_id, showSensitiveData)

    def update_info_card(self, db: Session, id: int, showSensitiveData=True):
        card_obj = db.query(CardSo).filter(CardSo.id == id).first()
        response_data = card.get(card_obj.search_id, showSensitiveData)
        response_card = response_data.data
        card_obj.PAN = response_card.pan
        card_obj.cvv = response_card.cvv
        card_obj.name = response_card.name
        card_obj.type = response_card.type
        card_obj.created_on = response_card.creation_time
        card_obj.expiry = response_card.expiration_date.date()
        card_obj.PAN = response_card.pan
        card_obj.status = response_card.status
        self.save_obj(db, card_obj)
        return card_obj

    def create_card(self, db: Session, user_id: int,
                    name: str, emboss_line4: str = None, type="VIRTUAL", card_label=None):
        wallet = db.query(WalletSo).filter(WalletSo.user_id == user_id).first()
        self.internal_transfer_to_safe(db, wallet.id, amount=self.card_price)
        if not card_label:
            user = db.query(self._user).filter(self._user.id == user_id).first()
            card_label = user.email


        response_data = card.create(
            owner_public_id=None,
            wallet_id=wallet.search_id,
            name=name,
            emboss_line4=emboss_line4,
            type=type,
            card_label=card_label)
        order = response_data.data

        if not order.is_valid or order.status != "PLACED":
            raise SoldoException("Error create_card", response_data.dict())
        result = {
            "wallet_id": wallet.id,
            "status": order.status,
            "category": "CARD"
        }
        self.__cache[order.id] = result
        return {order.id: result}

    def add_item_to_group(self, id: str, type="WALLET", groupId: str = None):
        if not groupId:
            groupId = self.settings.GROUP_ID
        return group.group_write(groupId, id, type)

    def get_order(self, order_id: str):
        return order.get(order_id)



from enum import Enum

from sqlalchemy import Column, Numeric, String, ForeignKey, Integer, Boolean, DateTime
from sqlalchemy.ext.declarative import as_declarative
from sqlalchemy.orm import relationship

from .base import CardBase, WalletBase, DateFixedMixin, CardType, CardStatus
from vc.db.base_class import BaseMixer


@as_declarative()
class SoldoBase(BaseMixer):
    pass


class WalletSo(WalletBase, DateFixedMixin, SoldoBase):
    __tablename__ = "soldo_wallet"
    blocked_balance = Column(Numeric, default=0)


class CardStatus(str, Enum):
    normal = "Normal"
    not_honor = "Do not honor"
    lost_card = "Lost card"
    stolen_card = "Stolen card"
    expired_card = "Expired card"
    restricted_card = "Restricted card"
    security_violation = "Security Violation"
    card_holder = "Cardholder to contact the issuer"
    destroyed = "Card Destroyed"
    pending = "pending"


class CardSo(CardBase, DateFixedMixin, SoldoBase):
    __tablename__ = "soldo_card"
    name = Column(String)
    wallet_id = Column(Integer, ForeignKey("soldo_wallet.id", ondelete="CASCADE"))
    wallet = relationship("WalletSo", backref="cards")
    type = Column(String, default=CardType.virtual.value)
    status = Column(String(20), default=CardStatus.pending.value)
    is_webhook = Column(Boolean, default=False)


class TransactionSo(SoldoBase):
    __tablename__ = "soldo_transaction"
    wallet_id = Column(Integer, ForeignKey("soldo_wallet.id", ondelete="CASCADE"))
    wallet = relationship("WalletSo", backref="transactions")
    search_id = Column(String)
    category = Column(String)
    amount = Column(Numeric)
    tx_amount_currency = Column(String(10))
    status = Column(String)
    date = Column(DateTime)
    update_time = Column(DateTime)
    fee_amount = Column(Numeric)
    fee_currency = Column(String(10))



import logging
from logging.handlers import TimedRotatingFileHandler
from vc import settings

def set_config(logger, filename: str = None):
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logger.setLevel(logging.DEBUG)
    # FileHandler

    fh = logging.StreamHandler()
    if settings.LOG_FILE_SOLDO:
        fh = TimedRotatingFileHandler(settings.LOG_FILE_SOLDO, when="W0", backupCount=2)
    fh.setFormatter(formatter)
    logger.addHandler(fh)


def generate_list_model_search_id(query_model):
    list_model_id = []
    list_model = {}
    for w in query_model:
        list_model_id.append(w.search_id)
        list_model[w.search_id] = w.id
    return list_model_id, list_model


import logging
from datetime import datetime

from sqlalchemy.orm import Session
from vc.libs.utils import set_config

from vc.models.soldo import WalletSo, CardSo


logger = logging.getLogger(__name__)
set_config(logger)


class EventMixer:
    format_date = "%Y-%m-%dT%H:%M:%SZ"

    # events functions

    def new_user(self, db: Session, **data):
        user = db.query(self._user).filter(self._user.email == data.get("email")).first()
        user.soldo_id = data.get("id")
        self.save_obj(db, user)

        if not data.get("groups"):
            self.add_item_to_group(user.soldo_id, type="USER")

        response_data = self.get_wallets(page_size=100)
        wallets = response_data.data.results
        wallets = list(filter(
            lambda w: w.primary_user_public_id == user.soldo_id,
            wallets))

        list_wallet_id = [w.search_id for w in user.wallet_soldo]

        wallets = list(filter(
            lambda w: w.id not in list_wallet_id,
            wallets
        ))

        for w in wallets:
            if w.currency_code in self.settings.currency:
                wallet = WalletSo(
                    user_id=user.id,
                    search_id=w.id,
                    balance=w.available_amount,
                    blocked_balance=w.blocked_amount,
                    currency=w.currency_code)
                if data.get("creation_time"):
                    wallet.created_on = datetime.strptime(data.get("creation_time"), self.format_date)
                self.save_obj(db, wallet)
        return user

    def wallet_created(self, db: Session, **data):
        """2021-07-02 12:56:37,564 - app.api.api_v1.endpoints.utils - DEBUG - b'{"event_type":"Wallet","event_name":"wallet_created",
        "data":
        {"id":"f0dd3007-838b-46e9-899c-3c526c71ee06","name":"test@test.com","currency_code":"USD","available_amount":0.00,"blocked_amount":0.00,"primary_user_type":"company","visible":true,"creation_time":"2021-07-02T09:56:37Z"}}'"""
        wallet = db.query(WalletSo).filter(WalletSo.search_id == data.get("id")).first()

        wallet.balance=data.get("available_amount"),
        wallet.blocked_balance=data.get("blocked_amount"),
        wallet.currency=data.get("currency_code")

        if data.get("creation_time"):
            wallet.created_on = datetime.strptime(data.get("creation_time"), self.format_date)
        self.save_obj(db, wallet)

        if not data.get("groups"):
            self.add_item_to_group(wallet.search_id)
        return wallet

    def store_order_completed(self, db, **data):
        cache = self.get_cache_by_key(data.get("id"))
        if cache:
            for i in data.get("items"):
                logger.info(cache)
                category = i.get("category")
                logger.debug(category)
                if category == "CARD":
                    logger.debug(i)
                    card = db.query(CardSo).filter(CardSo.search_id==i.get("id")).first()
                    if not card:
                        card = CardSo(search_id=i.get("id"), wallet_id=cache.get("wallet_id"), is_webhook=True)
                        self.save_obj(db, card)

                    card = self.update_info_card(db, card.id)
                    logger.debug(card)
                    logger.debug(card.__dict__)
                    self.add_item_to_group(card.search_id, type="CARD")
                    self.wallet_update_balance(db, card.wallet_id)
                    self.remove_cache(data.get("id"))


from typing import TypeVar, Optional

from pydantic.generics import GenericModel, Generic
from typing import Optional


T = TypeVar('T')


class ResponseData(GenericModel, Generic[T]):
    status_code: Optional[int]
    url: Optional[str]
    data: Optional[T]
    message: Optional[str]
    error_description: Optional[str]
    error: Optional[str]
    error_code: Optional[str]



class Settings(object):
    _config = {}

    def __init__(self, conf_dict):
        self._config = conf_dict

    def update_config(self, **kwargs):
        self._config.update(kwargs)

    def __getattr__(self, item):
        if item == "_config":
            return self._config
        # elif item == "update_config":
        #     return self.update_config
        return self._config.get(item)

    def __getitem__(self, item):
        return self._config[item]

    def __setitem__(self, key, value):
        if key == "_config":
            super().__setattr__(key, value)
        else:
            self._config[key] = value

    def __setattr__(self, key, value):
        self.__setitem__(key, value)


import os

LOG_FILE_SOLDO = os.getenv("VC_LOG_FILE_SOLDO", None)



from datetime import datetime
from .base_class import Base
from enum import Enum
from sqlalchemy import Column, Integer, ForeignKey, String, Date, DateTime, func, Numeric, Boolean
from sqlalchemy.orm import relationship


class StatusAccountVC(str, Enum):
    deactivated = "Deactivated"
    paid = "Paid"
    beta = "Beta"


class AccountVC(Base):
    __tablename__ = "vc_account"
    number = Column(String(100))
    balance = Column(Numeric, default=0)
    status = Column(String(15), default=StatusAccountVC.deactivated.value)

    user_id = Column(Integer, ForeignKey("user.id", ondelete="CASCADE"), nullable=False)
    # user = relationship("User", backref="vc_accounts")
    # campaigns = relationship("CampaignVC", back_populates="account")

    activated_on = Column(DateTime, index=True)
    created_on = Column(DateTime, default=datetime.now, server_default=func.now(), index=True)

    @property
    def is_active(self) -> bool:
        return bool(
            (self.status == StatusAccountVC.beta.value) or (self.status == StatusAccountVC.paid.value))

    def __repr__(self, **kwargs):
        return super().__repr__(f"id: {self.id}| user: {self.user.id}")


class CampaignType(str, Enum):
    # facebook = "facebook"
    # google = "google"
    facebook = "2"
    google = "1"
    prepaidcard = "3"


class CustomStatus(str, Enum):
    hidden = "Hidden"
    not_set = " "


class CampaignVC(Base):
    __tablename__ = "vc_campaign"
    number = Column(String(100))
    balance = Column(Numeric, default=0)
    account_id = Column(Integer, ForeignKey("vc_account.id", ondelete="CASCADE"), index=True, nullable=False)

    # account = relationship("AccountVC", back_populates="campaigns")

    type = Column(String, default=CampaignType.facebook.value)
    # cards = relationship("CardVC", back_populates="campaign")

    @property
    def count_cards(self):
        return len(self.cards)

    def __repr__(self, **kwargs):
        return super().__repr__(f"id {self.id} type - {self.type}")

    @property
    def sum_cards(self):
        return sum([card.balance for card in self.cards])


class StatusCard(str, Enum):
    pending = "PENDING"
    active = "ACT"
    block = "BLO"
    deleted = "DELETED"


class TypeCard(str, Enum):
    virtual_multi_use = "Virtual multi use"
    single_load_card = "Single load card"


class PaymentSystem(str, Enum):
    master_card = "MASTERCARD"
    visa = "VISA"


class CardVC(Base):
    __tablename__ = "vc_card"
    bin_id = Column(Integer)
    system_id = Column(Integer, index=True)
    reference = Column(String, index=True, unique=True, nullable=False)
    is_slave = Column(Boolean, default=False)
    campaign_vc_id = Column(Integer, ForeignKey("vc_campaign.id", ondelete="CASCADE"))
    # campaign = relationship("CampaignVC", back_populates="cards")
    balance = Column(Numeric, default=0)
    PAN = Column(String, index=True)
    cvv = Column(String(5))
    expiry = Column(Date)
    custom_status = Column(String)
    custom_comment = Column(String)
    custom_comment2 = Column(String)
    status = Column(String(20), default=StatusCard.pending.value)
    type = Column(String, default=TypeCard.virtual_multi_use.value)
    payment_system = Column(String, default=PaymentSystem.master_card.value)
    created_on = Column(DateTime, default=datetime.now, server_default=func.now(), index=True)


    def __repr__(self, **kwargs):
        return super().__repr__(f"id {self.id} | PAN: {self.PAN}")



from datetime import datetime
from enum import Enum

from sqlalchemy import DateTime, Column, func, Numeric, Integer, ForeignKey, String, Date
from sqlalchemy.ext.declarative import declared_attr

from vc.db.base_class import Base
from vc.db.models import StatusCard


class UserBase(Base):
    __tablename__ = "user"
    email = Column(String, nullable=False)
    first_name = Column(String)
    last_name = Column(String)
    job_title = Column(String)

    def __repr__(self, input_str=None):
        return super().__repr__(f"{self.email} {self.id}")


class WalletStatus(str, Enum):
    deactivated = "Deactivated"
    paid = "Paid"
    beta = "Beta"


class CardStatus(str, Enum):
    pending = "pending"
    active = "activated"
    block = "blocked"
    deleted = "deleted"


class CardType(str, Enum):
    virtual_multi_use = "Virtual multi use"
    virtual = "VIRTUAL"
    google = "GOOGLE_CARD"
    single_load_card = "Single load card"


class DateFixedMixin:
    created_on = Column(
        DateTime, default=datetime.now, server_default=func.now(), index=True
    )
    updated_on = Column(
        DateTime,
        default=datetime.now,
        onupdate=datetime.now,
        server_default=func.now(),
    )


class WalletBase:
    search_id = Column(String)
    status = Column(String(15), default=WalletStatus.deactivated.value)
    balance = Column(Numeric, nullable=False, default=0)
    currency = Column(String(10), default="EUR")

    @declared_attr
    def user_id(cls):
        return Column(Integer,
                      # ForeignKey("user.id", ondelete="CASCADE"),
                      nullable=False)


class CardBase:
    search_id = Column(String)
    PAN = Column(String, index=True)
    cvv = Column(String(5))
    expiry = Column(Date)
    status = Column(String(20), default=CardStatus.pending.value)
    type = Column(String, default=CardType.virtual_multi_use.value)



import logging
import requests

from vc.libs.decoratos import response_builder
from .schema_base import HeadersSoldoBase, HeadersSoldo, JWTData
from vc.libs.utils import set_config


logger = logging.getLogger(__name__)
set_config(logger)


class RequesterSoldoBase(object):
    default_authorize = HeadersSoldoBase
    advanced_authorize = HeadersSoldo
    auth2_data: JWTData

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @response_builder(data_schema=JWTData)
    def oauth_authorize(self):
        from vc.manager import Soldo
        api_path = f'/oauth/authorize'
        response_data = self.request(
            api_path, method='post',
            headers={"Content-Type": "application/x-www-form-urlencoded"},
            data={"client_id": Soldo.settings.CLIENT_ID, "client_secret": Soldo.settings.CLIENT_SECRET})
        return response_data

    def request(
            self, path: str, method: str = "get", *,
            headers: dict = None,
            params: str = None, data: dict = None,
            json: dict = None,
            **kwargs):
        from vc.manager import Soldo
        r = requests.request(url=Soldo.settings.API_URL + path, method=method, headers=headers, json=json, data=data, params=params, **kwargs)
        logger.debug(f"request: {r.request.url}, b: {str(r.request.body)} h: {str(r.request.headers)}")
        logger.debug(f"response: {r.text}, b:  h: {r.headers}")
        if r.status_code == 401 and "invalid_token" in r.text:
            response_data = self.oauth_authorize()
            Soldo.settings.ACCESS_TOKEN = response_data.data.access_token
            headers["Authorization"] = Soldo.settings.ACCESS_TOKEN
            return self.request(path, method,  headers=headers,
                                params=params, data=data, json=json)

        data = r.text
        try:
            data = r.json()
        except:
            pass
        finally:
            # fix long response
            if isinstance(data, str):
                data = data[:400]
        return data, r.status_code, r.url



from vc.libs.decoratos import response_builder
from .requesters.requester_base import RequesterSoldoBase
from .requesters.schemas import ResponseInfo, UserBase, Order, OrderItem, CardResponse, UserResponse, PaginateList, \
    WalletBase, ListRules, CardRule
import copy
from decimal import Decimal
from .requesters.utils import request_timestamp


class User(RequesterSoldoBase):

    @response_builder(data_schema=ResponseInfo)
    def whoami(self):
        api_path = f'/business/v2/ping/whoami'
        return self.request(api_path, method='get', headers=self.default_authorize().dict())

    @response_builder(data_schema=Order[OrderItem])
    def create(self, email: str, name: str, surname: str, custom_reference_id: str, job_title: str, **data):
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#add-user
        api_path = f"/business/v2/employees/"

        user = UserBase(
            request_timestamp=request_timestamp(),
            surname=surname,
            name=name,
            email=email,
            custom_reference_id=custom_reference_id,
            job_title=job_title,
            mobile_access=False,
            web_access=False,
            **data).dict(exclude_none=True)
        h = self.advanced_authorize(
            user, fields=("request_timestamp", "name", "surname", "mobile_access", "web_access"),
        ).dict(by_alias=True)
        return self.request(
            api_path, method='post',
            headers=h,
            json=user)

    @response_builder(data_schema=UserBase)
    def update(self, id, **data):
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#update-user-data
        api_path = f"/business/v2/employees/{id}"
        return self.request(
            api_path, method='put',
            headers=self.advanced_authorize(
                data, fields=("custom_reference_id", "job_title", "mobile_number", "mobile_prefix",
                              "email", "enable_mobile_credential", "enable_web_credential"),
            ).dict(by_alias=True),
            json=data)


class Wallets(RequesterSoldoBase):

    @response_builder(data_schema=ResponseInfo)
    def internal_transfer(self, amount: Decimal, toWalletId: str, fromWalletId: str, currencyCode: str= "EUR"):
        headers = self.advanced_authorize(
                dict(toWalletId=toWalletId, fromWalletId=fromWalletId, amount=amount, currencyCode=currencyCode),
                fields=("amount", "currencyCode", "fromWalletId", "toWalletId")
            )
        headers.Content_Type = "application/x-www-form-urlencoded"
        return self.request(
            f"/business/v2/wallets/internalTransfer/{fromWalletId}/{toWalletId}",
            method='PUT',
            headers=headers.dict(by_alias=True), data=dict(amount=amount, currencyCode=currencyCode))

    @response_builder(data_schema=WalletBase)
    def get(self, wallet_id: str):
        return self.request(
            f"/business/v2/wallets/{wallet_id}",
            method='get',
            headers=self.default_authorize().dict())

    @response_builder(data_schema=PaginateList[WalletBase])
    def search(self, page=0, page_size=50, **data):
        data.update(dict(
            s=page_size,
            p=page,
        ))
        return self.request(
            "/business/v2/wallets",
            params=data,
            headers=self.default_authorize().dict()
        )

    @response_builder(data_schema=Order[OrderItem])
    def create(self, owner_type, currency, name, **kwargs):
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#update-user-data
        # {
        # "id":"6a5e0887-eaf6-4f49-95cb-6b7f245d04f3",
        # "status":"PLACED","creation_time":"2021-06-29T10:42:06",
        # "last_update_time":"2021-06-29T10:42:06","is_valid":true,"total_paid_amount":0.000000,
        # "total_paid_currency":"EUR",
        # "items":[{"id":"87d95150-41f3-46a5-9501-ca1b368696ca","itemType":"WALLET","category":"WALLET"}]}
        api_path = f"/business/v2/wallets/"
        data = dict(request_timestamp=request_timestamp(), owner_type=owner_type, currency=currency, name=name,
                    **kwargs)
        h = self.advanced_authorize(
            data, fields=("request_timestamp", "owner_type", "currency", "name"),
        ).dict(by_alias=True)
        return self.request(
            api_path, method='post',
            headers=h, json=data)


class Card(RequesterSoldoBase):

    @response_builder(data_schema=Order[OrderItem])
    def create(self, wallet_id, owner_public_id: str = None, owner_type="company",
               type="VIRTUAL", name=None,
               emboss_line4=None, card_label="aff"):
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#update-user-data
        api_path = f"/business/v2/cards/"
        data = dict(request_timestamp=request_timestamp(), name=name, emboss_line4=emboss_line4, owner_type=owner_type,
                    owner_public_id=owner_public_id,
                    wallet_id=wallet_id, type=type, card_label=card_label)
        h = self.advanced_authorize(
            data
            , fields=("request_timestamp", "owner_type", "owner_public_id", "wallet_id"),
        ).dict(by_alias=True)
        return self.request(
            api_path, method='post',
            headers=h, json=data)

    @response_builder(data_schema=CardResponse)
    def get(self, card_id: str, showSensitiveData: str = None):
        return self.request(
            f"/business/v2/cards/{card_id}", method='get',
            params={"showSensitiveData": showSensitiveData},
            headers=self.default_authorize().dict())

    @response_builder(data_schema=ListRules)
    def get_card_rules(self, card_id: str):
        return self.request(
            f"/business/v2/cards/{card_id}/rules", method='get',
            headers=self.default_authorize().dict())

    @response_builder(data_schema=ListRules)
    def update_card_rule(self, card_id: str, name, enabled=None, amount=None):
        data = CardRule(
            name=name,
            enabled=enabled,
            amount=amount).dict(exclude_none=None)
        print(data)
        return self.request(
            f"/business/v2/cards/{card_id}/rules", method='put', json=data,
            headers=self.advanced_authorize(
                data, fields=("name", "amount", "enabled")).dict(by_alias=True))

    @response_builder(data_schema=PaginateList[CardResponse])
    def search(self, page=0, page_size=50, **data):
        data.update(dict(
            s=page_size,
            p=page,
        ))
        return self.request(
            "/business/v2/cards",
            params=data,
            headers=self.default_authorize().dict()
        )


class Order(RequesterSoldoBase):

    @response_builder(data_schema=Order)
    def get(self, order_id: str):
        api_path = f"/business/v2/orders/{order_id}"
        return self.request(
            api_path, method='get',
            headers=self.default_authorize().dict())


class Group(RequesterSoldoBase):

    # @response_builder(data_schema=UserBase)
    def group_write(self, groupId: str, id: str, type: str):
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#update-user-data
        data = dict(id=id, type=type)
        return self.request(
            f"/business/v2/groups/{groupId}/resource", method='post',
            headers=self.advanced_authorize(data).dict(by_alias=True), json=data)


class Transaction(RequesterSoldoBase):

    @response_builder(data_schema=PaginateList[ResponseInfo])
    def search(self, page=0, page_size=50, **params):
        params.update(dict(
            s=page_size,
            p=page,
        ))
        # http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#update-user-data
        return self.request(
            f"/business/v2/transactions", method='get',
            params=params,
            headers=self.advanced_authorize(
                params,
                fields=(
                "type", "publicId", "customReferenceId", "groupId", "fromDate", "toDate", "dateType", "category",
                "status", "tagId", "metadataId", "text")
            ).dict(by_alias=True))


group = Group()
user = User()
wallets = Wallets()
card = Card()
order = Order()
transaction = Transaction()
